// COPYRIGHT (c) 2002 PLOTNIX, INC. All rights reserved

package com.plotnix.jex;

import java.util.HashMap;

import com.plotnix.jex.adapters.InterpretingCompiler;


/**
 * A compiler translates an expression string into an executable form.
 * It validates the syntax of the expression and throws an exception
 * if the syntax is invalid.
 * <p>
 * In order to introduce a new compiler, create a class called
 * <code>"com.plotnix.jex.<i>language-prefix</i>.Compiler"</code> that 
 * inherits from this class.  Put it on the classpath. That's it. 
 * Jex will find thenew class by the prefix and invoke it to compile 
 * expressions in that language.
 *
 * @author Dmitri Plotnikov
 * @version $Revision$ $Date$
 */
public abstract class Compiler {

    /**
     * Serves as a placeholder for a missing compiler.  We don't want
     * to keep retrying to find the same compiler.
     */
    private static final Object KNOWN_MISSING_COMPILER = new Object();
    
    private static HashMap compilerMap = new HashMap();
    private static HashMap compilerImplMap = new HashMap();
    
    /**
     * Locates and returns a Compiler for the specified language.
     * If there is not such compiler, looks up an Interpreter
     * for this language. If one is found, returns an adapter Compiler
     * that makes an interpreter look like a Compiler.
     * If neither a Compiler nor Interpreter exists for the specified
     * language, returns null.
     * 
     * @param language is the name of an expression language (no colon
     *  at the end)
     * @return the Compiler for that language
     */
    public static final Compiler getCompiler(String language){
        Object compiler = compilerMap.get(language);
        if (compiler == null){
            compiler = getCompilerImplementation(language);
            if (compiler == null){
                Interpreter interpreter = 
                        Interpreter.getInterpreterImplementation(language);
                if (interpreter != null){
                    compiler = new InterpretingCompiler(interpreter);
                }
            }
            if (compiler == null){
                compiler = KNOWN_MISSING_COMPILER;
            }
            compilerMap.put(language, compiler);
        }        
        if (compiler == KNOWN_MISSING_COMPILER){
            return null;
        }
        return (Compiler) compiler;
    }
    
    /**
     * Locates and returns a Compiler for the specified language.
     * If no Compiler exists for the specified language, returns null.
     * 
     * @param language is the name of an expression language (no colon
     *  at the end)
     * @return the Compiler for that language
     */
    protected static final Compiler getCompilerImplementation(String language){
        Object compiler = compilerImplMap.get(language);
        if (compiler == null){
            Class clazz = null;
            try {
                clazz = Class.forName(
                    "com.plotnix.jex." + language + ".Compiler");
                compiler = (Compiler) clazz.newInstance();
            }
            catch (ClassNotFoundException ex){
                compiler = KNOWN_MISSING_COMPILER;
            }
            catch (Exception ex){
                // TBD: wrap
                throw new RuntimeException(ex.getMessage());
            }
            compilerImplMap.put(language, compiler);
        }        
        if (compiler == KNOWN_MISSING_COMPILER){
            return null;
        }
        return (Compiler) compiler;
    }
    
    /**
     * Compiles an expression intended for evaluation 
     * that produces a <i>value</i>.  For example,
     * <code>"2+2"</code> will be a valid argument for this method
     * for many expression languages.
     * <p>
     * May throw an UnsupportedOperationException.
     * 
     * @param expression 
     *  is an expression in the corresponding language
     * @return an executable object that can be used to compute 
     *  the value of this expression in a {@link Context Context}. 
     */ 
    public abstract Expression compileExpression(String expression);

    /**
     * Compiles an expression describing an iteration.
     * For example, <code>"addressBook//phone"</code> is a valid
     * argument for this method for an XPath compiler.
     * <p>
     * May throw an UnsupportedOperationException.
     * 
     * @param expression 
     *  is an iteration expression in the corresponding language
     * @return an executable object that can be used to 
     *  iterate over some values in a {@link Context Context}. 
     */ 
    public abstract IterationExpression compileIterationExpression(
            String expression);

    /**
     * Compiles an expression describing the left hand side
     * of an assignment.
     * 
     * For example, <code>"addressBook/contact[name="John Doe"]/phone"</code> 
     * is a valid argument for this method for an XPath compiler.
     * <p>
     * May throw an UnsupportedOperationException.
     * 
     * @param expression 
     *  is a variable expression in the corresponding language
     * @return an executable object that can be used to 
     *  modify a value in a {@link Context Context}. 
     */
    public abstract VariableExpression compileVariableExpression(
            String expression);            
}
