// COPYRIGHT (c) 2002 PLOTNIX, INC. All rights reserved

package com.plotnix.jex;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;


/**
 * Expressions are evaluated in a Context, which provides access
 * to shared pieces of the environment: variables, locale, expression
 * language compilers etc.
 *
 * @author Dmitri Plotnikov
 * @version $Revision$ $Date$
 */
public class Context {

    private Context parentContext;
    private String language;
    private boolean multilingual = true;
    private HashMap properties;
    private Variables variables;
    private Locale locale;
    
    /**
     * Creates a new Context
     */
    public Context(){
    }
    
    /**
     * Creates a new Context with the given parent Context. The child context
     * will provide access to the parent's variable, extension functions etc.
     * It can override and augment those.  The use of nested contexts
     * allows you to avoid re-configuring contexts all the time.
     * 
     * @param parentContext is the parent context
     */
    public Context(Context parentContext){
        this.parentContext = parentContext;
    }

    /**
     * Returns the Context's parent.
     * 
     * @return The parent Context or null if there is no parent.
     */
    public Context getParentContext(){
        return parentContext;
    }
        
    /**
     * Set the name of the expression language for this context. 
     * If the context is multilingual 
     * (see {@link #setMultilingual setMultilingual}), this language will 
     * function as the default - expressions in this language will 
     * not require a <code>"<i>lang</i>"</code> prefix.
     * 
     * @param language 
     *  is the name of an expression language. The compiler for
     *  this language should be available.
     */
    public void setExpressionLanguage(String language) {
        this.language = language;
    }
    
    /**
     * Returns the name of this context's (default) expression language.
     * 
     * @return the name of the language.
     */
    public String getExpressionLanguage() {
        return language;
    }
    
    /**
     * If set to true, this context will support multiple expression
     * languages. Expressions will start with the 
     * <code>"<i>lang:</i>"</code> prefix.  If the default expression
     * language has been specified, the expressions in that language
     * do not require a language prefix.
     * 
     * @param multilingual is the new setting
     */
    public void setMultilingual(boolean multilingual){
        this.multilingual = multilingual;
    }

    /**
     * Returns true if the Context is <code>multilingual</code>.
     * 
     * @see #setMultilingual
     */    
    public boolean isMultilingual(){
        return multilingual;
    }

    /**
     * Returns the compiler for the Context's expression language
     * or a multilingual compiler if the Context is multilingual.
     * <p>
     * Will throw an exception if the Context is not multilingual
     * and the expression language has not been specified.
     *
     * @see Compiler
     * @see MultilingualCompiler
     * 
     * @return The compiler
     */
    public Compiler getCompiler(){
        if (multilingual){
            return MultilingualCompiler.getMultilingualCompiler(language);
        }
        else if (language != null){
            return Compiler.getCompiler(language);
        }
        else {
            throw new RuntimeException(
               "No expression language has been specified for the context");
        }
    }
    
    /**
     * Returns the interpreter for the Context's expression language
     * or a multilingual interpreter if the Context is multilingual.
     * <p>
     * Will throw an exception if the Context is not multilingual
     * and the expression language has not been specified.
     *
     * @see Interpreter
     * @see MultilingualInterpreter
     * 
     * @return The interpreter
     */
    public Interpreter getInterpreter(){
        if (multilingual){
            return MultilingualInterpreter.getMultilingualInterpreter(language);
        }
        else if (language != null){
            return Interpreter.getInterpreter(language);
        }
        else {
            throw new RuntimeException(
                "No expression language has been specified for the context");
        }        
    }
    
    /**
     * Sets a configuration property that can be used by
     * expression language implementations. If the property
     * is language-specific, the name of the property should start
     * with "<i>lang.</i>" by convention.
     * 
     * @param property is the name of of the property
     * @param value is the value of the property
     */
    public void setProperty(String property, Object value){
        if (properties == null){
            properties = new HashMap();
        }
        properties.put(property, value);
    }
    
    /**
     * Returns the value of the specified property.
     * 
     * @param property is a property name
     * @return the value of the property
     */
    public Object getProperty(String property){
        if (properties != null){
            return properties.get(property);
        }
        return null;
    }
    
    /**
     * Returns the value of the specified property. If the property
     * is undefined on this context, tries to find it on the chain
     * of parent contexts.  Returns null if the property is undefined.
     * 
     * @param property is a property name
     * @return the value of the property
     */
    public Object findProperty(String property){
        if (parentContext != null){
            if (properties != null && properties.containsKey(property)){
                return properties.get(property);
            }
            return parentContext.getProperty(property);
        }
        if (properties != null){
            return properties.get(property);
        }
        return null;
    }
    
    /**
     * Install a custom implementation of a Variable pool. Initially
     * the context has the default variable pool, which is implemented
     * with a simple HashMap.
     * 
     * @param variables are custom Variables
     */
    public void setVariables(Variables variables){
        this.variables = variables;
    }

    /**
     * Returns the variable pool associated with the context. If no such
     * pool was specified during the construction of the context,
     * returns the default implementation of Variables which uses a HashMap.
     * 
     * @return variables - default or custom
     */
    public Variables getVariables(){
        if (variables == null){
            variables = new BasicVariables();
        }
        return variables;
    }

    /**
     * Set the locale for this context. By default, Jex uses
     * <code>Locale.getDefault()</code>
     * 
     * @param locale is the locale this context is associated with
     */
    public void setLocale(Locale locale){
        this.locale = locale;
    }

    /**
     * Returns the locale set with setLocale or, if none was set 
     * parent context's locale or Locale.getDefault().
     * 
     * @return The context's locale
     */
    public Locale getLocale(){
        if (locale == null){
            locale = Locale.getDefault();
        }
        if (parentContext != null){
            return parentContext.getLocale();
        }
        return locale;
    }
    
    /**
     * Invokes the context's interpreter to evaluate the supplied expression.
     * 
     * @see #getInterpreter()
     * 
     * @param expression is the expression that needs to be evaluated
     * @param arguments are argument of the expression. Can be null
     *  if the expression does not take any arguments.
     * @return the result of the evaluation of this expression
     */
    public Object evaluate(String expression, Object[] arguments){
        return getInterpreter().evaluate(this, expression, arguments);
    }

    /**
     * Evaluates the expression and casts the result to the supplied type.
     * 
     * @see #evaluate(String, Object[])
     */
    public Object evaluate(String expression, Object[] arguments, 
                           Class valueType){
        return getInterpreter().
            evaluate(this, expression, arguments, valueType);
    }

    /**
     * Invokes the context's interpreter to evaluate the supplied 
     * iteration expression.
     * 
     * @see #getInterpreter()
     * 
     * @param expression is the expression that needs to be evaluated
     * @param arguments are argument of the expression. Can be null
     *  if the expression does not take any arguments.
     * @return the result of the evaluation of this expression
     */
    public Iterator iterate(String expression, Object[] arguments){
        return getInterpreter().iterate(this, expression, arguments);
    }
    
    /**
     * Invokes the context's interpreter to modify the value
     * described by the supplied left-hand side expression.
     * Casts the type of the value if necessary.
     * 
     * @see #getInterpreter()
     * 
     * @param expression is the expression that needs to be evaluated
     * @param value is the new value
     * @param arguments are argument of the expression. Can be null
     *  if the expression does not take any arguments.
     */
    public void assign(String expression, Object value, Object[] arguments){
        getInterpreter().assign(this, expression, value, arguments);
    }    
    
    /**
     * The default implementation of Variables based on a HashMap.
     */ 
    private static class BasicVariables implements Variables {
        /**
         * Contains the values of declared variables
         */
        private HashMap vars = new HashMap();
    
        public boolean isDeclaredVariable(String varName){
            return vars.containsKey(varName);
        }
    
        public Object getVariable(String varName){
            // Note that a variable may be defined with a null value
    
            if (vars.containsKey(varName)){
                return vars.get(varName);
            }
    
            throw new IllegalArgumentException("No such variable: '" + 
                    varName + "'");
        }
    
        public void declareVariable(String varName, Object value){
            vars.put(varName, value);
        }
    
        public String toString(){
            return vars.toString();
        }
    }
}
