// COPYRIGHT (c) 2002 PLOTNIX, INC. All rights reserved

package com.plotnix.jex;


import java.util.HashMap;

/**
 * A MultilingualCompiler determines the language for each expression and
 * then compiles it using that language's compiler.
 * <p>
 * If the expression string passed to this compiler starts with 
 * <code>"<i>language</i>:"</code>, the compiler strips the prefix off and
 * passes the remainder of the expression to the compiler for 
 * the corresponding language. 
 * <p>The prefix is required if the MultilingualCompiler's default
 * language is null.
 * <p>If the default language is not null and the expression is missing
 * a prefix, or the prefix does not match an existing compiler,
 * the multilingual compiler passes the expression unchanged to the default 
 * language compiler.
 * 
 * @author Dmitri Plotnikov
 * @version $Revision$ $Date$
 */
public class MultilingualCompiler extends Compiler {

    private String defaultLanguage;
    private Compiler defaultCompiler;
    private static HashMap compilers = new HashMap();


    /**
     * Returns a MultilingualCompiler with the specified default language.
     * If the specified default language is not null, a compiler 
     * (or an interpreter) for that language should be available.
     * 
     * @param defaultLanguage is the default language or null
     * @return A MultilingualCompiler
     */
    public static final Compiler getMultilingualCompiler(
                String defaultLanguage){
        Compiler compiler = (Compiler) compilers.get(defaultLanguage);
        if (compiler == null){
            compiler = new MultilingualCompiler(defaultLanguage);
            compilers.put(defaultLanguage, compiler);
        }
        return compiler;
    }
    
    private MultilingualCompiler(String defaultLanguage){
        this.defaultLanguage = defaultLanguage;
        if (defaultLanguage != null){
           defaultCompiler = Compiler.getCompiler(defaultLanguage);
            if (defaultCompiler == null){
                throw new RuntimeException(
                    "Compiler not available for expression language " +
                    defaultLanguage);
            }
        }
    }
    
    /**
     * Compiles an expression intended for evaluation using the 
     * appropriate compiler.
     * If the expression has the syntax "<i>lang</i>:<i>expr</i>" and
     * a compiler is available for the language "<i>lang</i>", 
     * passes the expression to that compiler. Otherwise, passes the
     * whole unparsed expression to the default language compiler.
     * 
     * @param expression is the expression to be compiled with 
     *      an optional "lang:" prefix
     * 
     * @return the compiled expression
     */
    public Expression compileExpression(String expression) {        
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Compiler compiler = Compiler.getCompiler(language);
            if (compiler != null){
                String body = expression.substring(inx + 1);
                return compiler.compileExpression(body);
            }
        }
        if (defaultCompiler != null){
            return defaultCompiler.compileExpression(expression);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }

    /**
     * Compiles an expression that describes an iteration.
     * 
     * @see #compileExpression(String)
     */
    public IterationExpression compileIterationExpression(String expression) {
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Compiler compiler = Compiler.getCompiler(language);
            if (compiler != null){
                String body = expression.substring(inx + 1);
                return compiler.compileIterationExpression(body);
            }
        }
        if (defaultCompiler != null){
            return defaultCompiler.compileIterationExpression(expression);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }

    /**
     * Compiles an expression intended to serve as the left hand
     * side of an assignment.
     * 
     * @see #compileExpression(String)
     */
    public VariableExpression compileVariableExpression(String expression) {
        int inx = expression.indexOf(':');
        if (inx != -1){
            String language = expression.substring(0, inx);
            Compiler compiler = Compiler.getCompiler(language);
            if (compiler != null){
                String body = expression.substring(inx + 1);
                return compiler.compileVariableExpression(body);
            }
        }
        if (defaultCompiler != null){
            return defaultCompiler.compileVariableExpression(expression);
        }
        else {
            throw new RuntimeException(
                "The expression does not have a 'lang:' prefix and " +
                "no default language has been specified");
        }
    }
}
