// COPYRIGHT (c) 2002 PLOTNIX, INC. All rights reserved

package com.plotnix.jex.javascript;

import org.mozilla.javascript.NativeJavaObject;
import org.mozilla.javascript.Script;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;

import com.plotnix.jex.Context;
import com.plotnix.jex.Variables;

/**
 * 
 *
 * @author Dmitri Plotnikov
 * @version $Revision$ $Date$
 */
public abstract class JavaScriptExpression {

    /**
     * Jex stores the Global Scope for evaluation of JavaScript
     * expressions in the root Jex Context as a property named
     * JS_GLOBAL_SCOPE_PROPERTY
     */
    public static final String JS_GLOBAL_SCOPE_PROPERTY = 
                "javascript.GLOBAL_SCOPE";
                
    /**
     * Jex stores the Local Scope for evaluation of JavaScript
     * expressions in the current Jex Context as a property named
     * JS_SCOPE_PROPERTY
     */
    public static final String JS_SCOPE_PROPERTY = "javascript.SCOPE";

    /**
     * By default, Jex does not allow assignment to undeclared variables.
     * To relax this restriction, define a property called 
     * JS_AUTODECLARE ("javascript.autodeclare") in the context that
     * will host the new variables.
     */
    public static final String JS_AUTODECLARE = "javascript.autodeclare";    
    
    /**
     * Name of the pseudo-variable used by JavaScriptVariableExpression
     * to hold the value to be assigned.
     */
    protected static final String JS_PSEUDO_VARIABLE = "__value__";
    
    private Script script;
    
    /**
     * @param script is a compiled JavaScript to be associated with
     *  this expression
     */
    public JavaScriptExpression(Script script){
        this.script = script;
    }

    /**
     * @return the compiled JavaScript 
     */
    public Script getScript(){
        return script;
    }
    
    /**
     * Creates a new scope for the evaluation of an expression. 
     * Makes the new scope a child of the Scriptable associated 
     * with the Context.
     * 
     * @param context is a Jex Context
     * @param jsContext is a javascript context for this operation
     * @param arguments should have the length of zero or one. The
     *    first argument of the array (or null) is used as the context bean.
     * @return Local scope
     */
    protected Scriptable getLocalScope(Context context, 
            org.mozilla.javascript.Context jsContext, 
            Object[] arguments){
        return getLocalScope(context, jsContext, arguments, false, null);
    }
            
    /**
     * Creates a new scope for the evaluation of an expression. 
     * Makes the new scope a child of the Scriptable associated 
     * with the Context.  Declares a pseudo-variable in the new context
     * that is used by JavaScriptVariableExpression as a holder for
     * the value to be assigned.
     * 
     * @param context is a Jex Context
     * @param jsContext is a javascript context for this operation
     * @param arguments should have the length of zero or one. The
     *    first argument of the array (or null) is used as the context bean.
     * @param value is the value assigned to the pseudo-variable
     * @return Local scope
     */
    protected Scriptable getLocalScope(Context context, 
            org.mozilla.javascript.Context jsContext, 
            Object[] arguments, Object value){
        return getLocalScope(context, jsContext, arguments, true, value);
    }
            
    private Scriptable getLocalScope(Context context, 
            org.mozilla.javascript.Context jsContext, 
            Object[] arguments,
            boolean supportValue, Object value){
        Object root;
        if (arguments == null || arguments.length == 0){
            root = null;
        }
        else if (arguments.length == 1){
            root = arguments[0];
        }
        else {
            throw new RuntimeException("Too many arguments");
        }

        Scriptable parentScope = getContextScope(context, jsContext);
        if (root != null){
            Scriptable njo = 
                new NativeJavaObject(parentScope, root, root.getClass());
            njo.setPrototype(parentScope);
            parentScope = njo;
        }
        
        Scriptable localScope = parentScope;
        if (supportValue){
            localScope = new PseudoVariableScope(value);
            localScope.setPrototype(parentScope);
            localScope.setParentScope(null);
        }
        return localScope;
    }
    
    /**
     * Creates a JavaScript scope for this JexContext. Provides access
     * to the context variables.  Caches the JavaScript scope in the
     * context itself.
     * 
     * @param context is a Jex Context
     * @param jsContext is a javascript context for this operation
     * @return Context scope
     */
    protected Scriptable getContextScope(Context context,
            org.mozilla.javascript.Context jsContext){
        Scriptable scope = (Scriptable) context.getProperty(JS_SCOPE_PROPERTY);
        if (scope == null){
            scope = new ContextScope(context);
            scope.setPrototype(getGlobalScope(context, jsContext));
            scope.setParentScope(null);
            context.setProperty(JS_SCOPE_PROPERTY, scope);
        }   
        return scope;
    }
    
    /**
     * Creates a global JavaScript scope. Caches the scope in the
     * root Jex context.
     * 
     * @param context is a Jex Context
     * @param jsContext is a javascript context for this operation
     * @return Global scope
     */
    protected Scriptable getGlobalScope(Context context,
            org.mozilla.javascript.Context jsContext){        
        Scriptable scope = 
            (Scriptable) context.getProperty(JS_GLOBAL_SCOPE_PROPERTY);
        if (scope == null){
            scope = jsContext.initStandardObjects(null);
            Context root = context;
            while (root.getParentContext() != null){
                root = root.getParentContext();
            }
            root.setProperty(JS_GLOBAL_SCOPE_PROPERTY, scope);
        }
        return scope;
    }
    
    /**
     * A JavaScript scriptable that provides access to Jex Context
     * variables.
     */
    private static class ContextScope extends ScriptableObject {
        private Context context;
        
        public ContextScope(Context context){
            this.context = context;
        }
        
        public String getClassName(){
            return "JexContext";                
        }
        
        public Object get(String name, Scriptable start){
            Object result = super.get(name, start);
            if (result == NOT_FOUND) {
                Context varCtx = findVariableContext(context, name);
                if (varCtx != null){
                    result = varCtx.getVariables().getVariable(name);
                }
            }
            return result;
        }
        
        
        public void put(String name, Scriptable start, Object value){
            Context varCtx = findVariableContext(context, name);
            if (varCtx == null){
                varCtx = findAutodeclareContext(context);
            }
            
            if (varCtx != null){
                varCtx.getVariables().declareVariable(name, value);
            }            
            else {               
                throw new RuntimeException("Unknown variable: " + name);
            }            
        }
        
        private Context findVariableContext(Context context, String varName){
            while (context != null){
                Variables vars = context.getVariables();
                if (vars.isDeclaredVariable(varName)){
                    return context;                    
                }
                context = context.getParentContext();
            }
            return null;
        }
        
        private Context findAutodeclareContext(Context context){
            while (context != null){
                if (context.getProperty(JS_AUTODECLARE) != null){
                    return context;
                }
                context = context.getParentContext();
            }
            return null;
        }
    };    

    /**
     * A JavaScript scriptable that provides access to Jex Context
     * variables.
     */
    private static class PseudoVariableScope extends ScriptableObject {
        private Object value;
        
        public PseudoVariableScope(Object value){
            this.value = value;
        }
        
        public String getClassName(){
            return "JexContext";                
        }
        
        public Object get(String name, Scriptable start){
            if (name.equals(JS_PSEUDO_VARIABLE)){
                return value;
            }
            return super.get(name, start);
        }
                
        public void put(String name, Scriptable start, Object value){
            getPrototype().put(name, start, value);
        }
    };    
}
